import numba, numpy as np , math 

# Gravatational constant 
G = 6.67408e-11

#######################################
#          Keplers quation            #
#######################################

@numba.njit(fastmath=True)
def kepler(M, E, e) : return M - E - e*math.sin(E) 

@numba.njit(fastmath=True)
def kdepler(E, e) : return -1 + e*math.cos(E) 


@numba.njit(fastmath=True)
def fmod(a, b) : return a % abs(b)

#######################################
#        Eccentric anomaly            #
#######################################
@numba.njit(fastmath=True)
def getEccentricAnomaly(M, e, E_tol):
    m =  fmod(M , 2*math.pi) # should be fmod
    f1 = fmod(m , (2*math.pi) + e*math.sin(m)) + e*e*math.sin(2.0*m)/2.0  # should be fmod around 
    test  =1.0 
    e1 = 0.
    e0 = 1.0 
    while (test > E_tol):
        e0 = e1 
        e1 = e0 + (m-(e0 - e*math.sin(e0)))/(1.0 - e*math.cos(e0))
        test = abs(e1-e0)

    if (e1 < 0) : e1 = e1 + 2*math.pi 
    return e1 
 
#######################################
#      Time of periastron passage     #
#######################################
@numba.njit(fastmath=True)
def t_ecl_to_peri(t_ecl, e, w, incl, p_sid):
    # Define variables used
    efac  = 1.0 - e*2       # check if 2 or power
    sin2i = math.sin(incl)**2.

    # Value of theta for i=90 degrees
    ee = 0.
    theta_0 = (math.pi/2.) - w       # True anomaly at superior conjunction

    if (theta_0 == math.pi) :  ee = math.pi
    else :  ee =  2.0 * math.atan(math.sqrt((1.-e)/(1.0+e)) * math.tan(theta_0/2.0))

    eta = ee - e*math.sin(ee)
    delta_t = eta*p_sid/(2*math.pi)
    return t_ecl  - delta_t 

#######################################
#      Calculate the true anomaly     #
#######################################
@numba.njit(fastmath=True)
def getTrueAnomaly(time, e, w, period,  t_zero, incl, E_tol ):

    # Calcualte the mean anomaly
    M = 2*math.pi*((time -  t_ecl_to_peri(t_zero, e, w, incl, period)  )/period % 1.)

    # Calculate the eccentric anomaly
    E = getEccentricAnomaly(M, e, E_tol)

    # Now return the true anomaly
    return 2.*math.atan(math.sqrt((1.+e)/(1.-e))*math.tan(E/2.)) 

#######################################
#  Calculate the projected seperaton  #
#######################################
@numba.njit(fastmath=True)
def get_z(nu, e, incl, w, radius_1) : return (1-e**2) * math.sqrt( 1.0 - math.sin(incl)**2  *  math.sin(nu + w)**2) / (1 + e*math.sin(nu)) /radius_1

@numba.njit(fastmath=True)
def get_z_(nu, z) : return get_z(nu, z[0], z[1], z[2], z[3]) # for brent

#######################################
#  Calculate the projected seperaton  #
#######################################
@numba.njit(fastmath=True)
def getProjectedPosition(nu, w, incl) : return math.sin(nu + w)*math.sin(incl) 


#######################################
#     Calculate the mass function     #
#######################################
@numba.njit(fastmath=True)
def mass_function_1(e, P, K1) : return (1-e*e) **1.5*P*86400.1* (K1*1000)**3/(2*math.pi*G*1.989e30)

@numba.njit(fastmath=True)
def mass_function_1_(M2, z) : return ((M2*math.sin(z[1]))**3 / ( (z[0] + M2)**2)) - mass_function_1(z[2], z[3], z[4]) 


import math, numba 


@numba.njit
def Flux_drop_uniform( z, k, SBR):
    if (z >= 1. + k) or ((z >= 1.) and  (z <= k - 1.)) :  return 0.0           # total eclipse of the star
    elif  (z <= 1. - k)                                :  return - SBR*k*k   # planet is fully in transit		
    else :                                                                     #  planet is crossing the limb
        kap1 = math.acos(min((1. - k*k + z*z)/2./z, 1.))
        kap0 = math.acos(min((k*k + z*z - 1.)/2./k/z, 1.))
        return - SBR*  (k*k*kap0 + kap1 - 0.5*math.sqrt(max(4.*z*z - math.pow(1. + z*z - k*k, 2.), 0.)))/math.pi 


'''
import numpy as np 
import matplotlib.pyplot as plt 
z = np.linspace(-2,2,1000) 
_z = np.abs(z)
k = 0.1
SBR = 0.4 


for i in range(len(z)) : plt.scatter(z[i], Flux_drop_uniform( _z[i], k, SBR)) 
plt.show()
'''




# Python imports
import numba, numba.cuda, numpy as np , math 

import matplotlib.pyplot as plt 

# bruce imports
from .kepler import getTrueAnomaly

@numba.njit(fastmath=True)
def _rv( time, RV, RV_err, J,
        t_zero, period, 
        K1, fs, fc,
        V0, dV0,
        incl,
        E_tol, 
        loglike_switch):

    # Unpack and convert (assume incl in radians)
    w = math.atan2(fs, fc)
    e = fs**2 + fc**2

    loglike=0.
    for i in range(time.shape[0]):
        nu = getTrueAnomaly(time[i], e, w, period,  t_zero, incl, E_tol ) 
        model = K1*(e*math.cos(w) + math.cos(nu + w)) + V0 + dV0*(time[i] - t_zero) 
        if loglike_switch : 
            wt = 1.0 / (RV_err[i]**2 + J**2)
            loglike += -0.5*((RV[i] - model)**2*wt - math.log(wt))
        else : RV[i] = model 

    return loglike


@numba.cuda.jit
def kernel_rv(time, RV, RV_err, J,
        t_zero, period, 
        K1, fs, fc,
        V0, dV0,
        incl,
        E_tol, 
        loglike):

    # Unpack and convert (assume incl in radians)
    w = math.atan2(fs, fc)
    e = fs**2 + fc**2

    i = numba.cuda.grid(1)

    nu = getTrueAnomaly(time[i], e, w, period,  t_zero, incl, E_tol ) 
    model = K1*(e*math.cos(w) + math.cos(nu + w)) + V0 + dV0*(time[i] - t_zero) 
    wt = 1.0 / (RV_err[i]**2 + J**2)
    loglike[i] = -0.5*((RV[i] - model)**2*wt - math.log(wt))        


@numba.cuda.reduce
def sum_reduce(a, b):
    return a + b


def rv( time, RV = np.zeros(1), RV_err=np.zeros(1), J=0,
        t_zero=0., period=1., 
        K1=10., fs=0., fc=0.,
        V0=10., dV0=0.,
        incl=90.,
        E_tol=1e-5, 
        gpu=0, loglike=np.zeros(1), blocks = 10, threads_per_block = 512):

    # Convert inclination 
    incl = np.pi * incl/180.
    if not gpu:
        # First, let's see if we need loglike or not!
        if RV_err[0]==0 : loglike_switch = 0
        else            : loglike_switch = 1

        # Now, let's initiase the arrays, if needed
        if not loglike_switch : RV = np.empty_like(time) 

        # Now make the call
        loglike = _rv( time, RV, RV_err, J,
            t_zero, period, 
            K1, fs, fc,
            V0, dV0,
            incl,
            E_tol, 
            loglike_switch) 

        if loglike_switch : return loglike 
        else              : return RV 
    

    if gpu:
        # Loglike ony supported 
        # assumeing loglike is array

        ## Call the kernel to populate loglike 
        kernel_rv[blocks, threads_per_block](time, RV, RV_err, J,
            t_zero, period, 
            K1, fs, fc,
            V0, dV0,
            incl,
            E_tol, 
            loglike)
        
        # let's synchronise to ensure it's finished
        numba.cuda.synchronize() 

        # Now reduce the loglike array
        return sum_reduce(loglike)


'''
time = np.linspace(2450000, 2450001, 100000)
RV = rv(time)
RV = np.random.normal(RV, 1) 
RV_err = np.random.uniform(0.5,1.5, RV.shape[0])
J = 0.2 

CPU_loglike = rv(time, RV, RV_err, J=J) 
print('CPU loglike = ', CPU_loglike)

# Now copy stuff over to GPU for compliance 
d_time = numba.cuda.to_device(time)
d_RV = numba.cuda.to_device(RV)
d_RV_err = numba.cuda.to_device(RV_err)
loglike = np.empty_like(time) 
d_loglike = numba.cuda.to_device(loglike)

threads_per_block=512 
blocks = int(np.ceil(time.shape[0]/threads_per_block))
GPU_loglike = rv(d_time, d_RV, d_RV_err, J=J, gpu=1, loglike=d_loglike, blocks=blocks, threads_per_block=threads_per_block) 
print('GPU loglike = ', GPU_loglike)



# Time it 
timeit GPU_loglike = rv(d_time, d_RV, d_RV_err, J=J, gpu=1, loglike=d_loglike, blocks=blocks, threads_per_block=threads_per_block) 



plt.plot(time,RV)

plt.figure() 
K1 = np.linspace(5,15, 100)
for i in range(len(K1)) : plt.scatter(K1[i] , rv(time, RV, RV_err, J=J, K1=K1[i]))
plt.show()
'''


import numba, math 


@numba.njit
def Fellipsoidal(nu, q, radius_1, incl, u, y):
    # SUrray of https:#books.google.co.uk/books?id=ngtmDwAAQBAJ&pg=PA239&lpg=PA239&dq=ellipsoidal+variation+approximation+binary+star&source=bl&ots=swiO_JQdIR&sig=ACfU3U0HVtS8G37Z7EbdjDymUqICD36FgA&hl=en&sa=X&ved=2ahUKEwiO1tH9ud7hAhWDaFAKHRVoASIQ6AEwC3oECAkQAQ#v=onepage&q=ellipsoidal%20variation%20approximation%20binary%20star&f=false
    # q - mass ratio
    # radius_1 - R*/a 
    # incl - orbital inclination 
    # u - linear limb-darkening coefficient
    # y - gravitational darkening coefficient 
    # nu - true anomaly 
    #nu /= 2
    # true anomaly goes from -pi to pi 
    # to convert to phase,we could ass pi and divide by two pi
    alpha1 = ((y+2.)/(y+1.))*25.*u / (24.*(15. + u))
    alpha2 = (y+1.)*(3.*(15.+u))/(20.*(3.-u))
    Ae = alpha2*q*math.pow(radius_1,3)*math.pow(math.sin(incl),2) # Amplitude of ellipsoidal variation

    # Harmonic coeefs 
    f1 = 3*alpha1*radius_1*(5*math.pow(math.sin(incl),2) - 4)/math.sin(incl)
    f2 = 5*alpha1*radius_1*math.sin(incl)

    # Now return the variation 
    return -Ae*( math.cos(2*nu)    +    f1*math.cos(nu)      +      f2*math.cos(3*nu) ) 

'''
import numpy as np 
import matplotlib.pyplot as plt 

nu = np.linspace(0, 2*np.pi, 100) 
q = 0.4 
radius_1 = 0.2 
k = 0.2 
incl = math.pi/2 
u = 0.5
y = 0.4 

for i in nu : plt.scatter(i, Fellipsoidal(i, q, radius_1, incl, u, y) ) 
plt.show()
'''


import math, numba 


@numba.njit 
def clip(a, b, c) : 
    if a < b : return b 
    if a > c : return c 
    return a

@numba.njit
def ellpic_bulirsch( n,  k):
    kc = math.sqrt(1.-k*k)
    p = math.sqrt(n + 1.)
    m0 = 1.
    c = 1.
    d = 1./p
    e = kc
    nit = 0
    while(nit < 10000):
        f = c
        c = d/p + c
        g = e/p
        d = 2.*(f*g + d)
        p = g + p
        g = m0
        m0 = kc + m0
        if(math.fabs(1.-kc/g) > 1.0e-8):
            kc = 2.*math.sqrt(e)
            e = kc*m0
        else : return 0.5*math.pi*(c*m0+d)/(m0*(m0+p))
        nit+=1
    return 0.

@numba.njit
def ellec(k):
    # Computes polynomial approximation for the complete elliptic
    # integral of the first kind (Hasting's approximation):
    m1 = 1.0 - k*k
    a1 = 0.44325141463
    a2 = 0.06260601220
    a3 = 0.04757383546
    a4 = 0.01736506451
    b1 = 0.24998368310
    b2 = 0.09200180037
    b3 = 0.04069697526
    b4 = 0.00526449639
    ee1 = 1.0 + m1*(a1 + m1*(a2 + m1*(a3 + m1*a4)))
    ee2 = m1*(b1 + m1*(b2 + m1*(b3 + m1*b4)))*math.log(1.0/m1)
    ellec = ee1 + ee2
    return ellec

@numba.njit
def ellk( k):
    # Computes polynomial approximation for the complete elliptic
    # integral of the second kind (Hasting's approximation):
    m1 = 1.0 - k*k
    a0 = 1.38629436112
    a1 = 0.09666344259
    a2 = 0.03590092383
    a3 = 0.03742563713
    a4 = 0.01451196212
    b0 = 0.5
    b1 = 0.12498593597
    b2 = 0.06880248576
    b3 = 0.03328355346
    b4 = 0.00441787012
    ek1 = a0 + m1*(a1 + m1*(a2 + m1*(a3 + m1*a4)))
    ek2 = (b0 + m1*(b1 + m1*(b2 + m1*(b3 + m1*b4))))*math.log(m1)
    ellk = ek1 - ek2
    return ellk



@numba.njit
def Flux_drop_analytical_quadratic( d,  p,  c1,  c2,  tol):
    '''
    Calculate the analytical flux drop from the quadratic limb-darkening law.

    Parameters
    d : double
        Projected seperation of centers in units of stellar radii.
    p : double
        Ratio of the radii.
    c : double
        The first power-2 coefficient.
    a : double
        The second power-2 coefficient.
    f : double
        The flux from which to drop light from.
    eps : double
        Factor (1e-9)
    '''

    kap0 = 0.0
    kap1 = 0.0
    omega = 1.0 - c1/3.0 - c2/6.0

    # allow for negative impact parameters
    d = math.fabs(d)

    # check the corner cases
    if(math.fabs(p - d) < tol) : d = p
    
    if(math.fabs(p - 1.0 - d) < tol) : d = p - 1.0
    
    if(math.fabs(1.0 - p - d) < tol) : d = 1.0 - p
    
    if(d < tol) : d = 0.0
    

    x1 = math.pow((p - d), 2.0)
    x2 = math.pow((p + d), 2.0)
    x3 = p*p - d*d

    #source is unocculted:
    if(d >= 1.0 + p) : return 0.
    
    #source is completely occulted:
    if(p >= 1.0) and (d <= p - 1.0):
        lambdad = 0.0
        etad = 0.5        #error in Fortran code corrected here, following Jason Eastman's python code
        lambdae = 1.0
        return - ((1.0 - c1 - 2.0*c2)*lambdae + (c1 + 2.0*c2)*(lambdad + 2.0/3.0) + c2*etad)/omega
    
    #source is partly occulted and occulting object crosses the limb:
    if(d >= math.fabs(1.0 - p)) and (d <= 1.0 + p):
        kap1 = math.acos(   clip(min((1.0 - p*p + d*d)/2.0/d, 1.0),-1,1)   )
        #print(p, d, (p*p + d*d - 1.0)/2.0/p/d )
        kap0 = math.acos(   clip(min((p*p + d*d - 1.0)/2.0/p/d, 1.0),-1,1) )
        lambdae = p*p*kap0 + kap1
        lambdae = (lambdae - 0.50*math.sqrt(max(4.0*d*d - math.pow((1.0 + d*d - p*p), 2.0), 0.0)))/math.pi
    

    #edge of the occulting star lies at the origin
    if(d == p):
        if(d < 0.5):
            q = 2.0*p
            Kk = ellk(q)
            Ek = ellec(q)
            lambdad = 1.0/3.0 + 2.0/9.0/math.pi*(4.0*(2.0*p*p - 1.0)*Ek + (1.0 - 4.0*p*p)*Kk)
            etad = p*p/2.0*(p*p + 2.0*d*d)
            return - ((1.0 - c1 - 2.0*c2)*lambdae + (c1 + 2.0*c2)*lambdad + c2*etad)/omega

        elif(d > 0.5):
            q = 0.5/p
            Kk = ellk(q)
            Ek = ellec(q)
            lambdad = 1.0/3.0 + 16.0*p/9.0/math.pi*(2.0*p*p - 1.0)*Ek -  \
                (32.0*math.pow(p, 4.0) - 20.0*p*p + 3.0)/9.0/math.pi/p*Kk
            etad = 1.0/2.0/math.pi*(kap1 + p*p*(p*p + 2.0*d*d)*kap0 -  \
                (1.0 + 5.0*p*p + d*d)/4.0*math.sqrt((1.0 - x1)*(x2 - 1.0)))

        else:
            lambdad = 1.0/3.0 - 4.0/math.pi/9.0
            etad = 3.0/32.0
            return - ((1.0 - c1 - 2.0*c2)*lambdae + (c1 + 2.0*c2)*lambdad + c2*etad)/omega
        

        return - ((1.0 - c1 - 2.0*c2)*lambdae + (c1 + 2.0*c2)*lambdad + c2*etad)/omega
    
    #occulting star partly occults the source and crosses the limb:
    #if((d > 0.5 + math.fabs(p  - 0.5) && d < 1.0 + p) || (p > 0.5 && d > math.fabs(1.0 - p)*1.0001 \
    #&& d < p))  #the factor of 1.0001 is from the Mandel/Agol Fortran routine, but gave bad output for d near math.fabs(1-p)
    if ( (d > 0.5 + math.fabs(p  - 0.5)) and (d < 1.0 + p) ) or ( (p > 0.5) and (d > math.fabs(1.0 - p)) and (d < p)):
        q = math.sqrt((1.0 - x1)/4.0/d/p)
        Kk = ellk(q)
        Ek = ellec(q)
        n = 1.0/x1 - 1.0
        Pk = ellpic_bulirsch(n, q)
        lambdad = 1.0/9.0/math.pi/math.sqrt(p*d)*(((1.0 - x2)*(2.0*x2 +  \
                x1 - 3.0) - 3.0*x3*(x2 - 2.0))*Kk + 4.0*p*d*(d*d +  \
                7.0*p*p - 4.0)*Ek - 3.0*x3/x1*Pk)
        if(d < p) : lambdad += 2.0/3.0
        etad = 1.0/2.0/math.pi*(kap1 + p*p*(p*p + 2.0*d*d)*kap0 -  \
            (1.0 + 5.0*p*p + d*d)/4.0*math.sqrt((1.0 - x1)*(x2 - 1.0)))
        return - ((1.0 - c1 - 2.0*c2)*lambdae + (c1 + 2.0*c2)*lambdad + c2*etad)/omega
    
    #occulting star transits the source:
    if (p <= 1.0) and (d <= (1.0 - p)):
        etad = p*p/2.0*(p*p + 2.0*d*d)
        lambdae = p*p

        q = math.sqrt((x2 - x1)/(1.0 - x1))
        Kk = ellk(q)
        Ek = ellec(q)
        n = x2/x1 - 1.0
        Pk = ellpic_bulirsch(n, q)

        lambdad = 2.0/9.0/math.pi/math.sqrt(1.0 - x1)*((1.0 - 5.0*d*d + p*p +  \
                x3*x3)*Kk + (1.0 - x1)*(d*d + 7.0*p*p - 4.0)*Ek - 3.0*x3/x1*Pk)

        # edge of planet hits edge of star
        if(math.fabs(p + d - 1.0) <= tol):        
            lambdad = 2.0/3.0/math.pi*math.acos( clip(1.0 - 2.0*p,-1,1 )) - 4.0/9.0/math.pi* \
                        math.sqrt(p*(1.0 - p))*(3.0 + 2.0*p - 8.0*p*p)
        
        if(d < p) : lambdad += 2.0/3.0
    
    return - ((1.0 - c1 - 2.0*c2)*lambdae + (c1 + 2.0*c2)*lambdad + c2*etad)/omega


'''
import numpy as np 
import matplotlib.pyplot as plt 
z = np.linspace(-2,2,1000)
z_ = np.abs(z) 

k = 0.1 
c1 = 0.5
c2 = 0.2
tol = 1e-8  


for i in range(len(z)) : plt.scatter(z[i], Flux_drop_analytical_quadratic( z_[i],  k,  c1,  c2,  tol), c='k', s=10)
plt.show() 
'''


from setuptools import setup, Extension

setup(
    name = 'bruce-package',
    version = '0.5',
    description = 'GPU-accelerated binary star model',
    url = None,
    author = 'Samuel Gill et al',
    author_email = 'samuel.gill@warwick.ac.uk',
    license = 'GNU',
    packages=['bruce','bruce/binarystar'],
    scripts=['ngtsfit/ngtsfit'],
    install_requires=['celerite', 'numba', 'numpy', 'emcee']
)